"""Utility functions for working with Terraform modules"""
import concurrent.futures
import os
from typing import Set, Tuple

import hcl2
from networkx import DiGraph


def get_module_usage_graph(root_directory: str) -> DiGraph:
    """
    Recursively scan a directory with terraform config files and return what modules are being used
    :param root_directory: Directory to scan
    :return: Graph of module path and lists of directories that depend on each module
    """
    graph = DiGraph()
    future_list = []
    with concurrent.futures.ThreadPoolExecutor(max_workers=50) as executor:
        # pylint: disable=unused-variable
        for current_dir, dirs, files in os.walk(root_directory, followlinks=True):
            if '.terraform' in current_dir or '.git' in current_dir:
                continue

            for file in files:
                if not file.endswith('.tf'):
                    continue
                future = executor.submit(_get_modules_for_file, current_dir, file)
                future_list.append(future)

        for future in concurrent.futures.as_completed(future_list):
            directory, modules = future.result()
            for mod in modules:
                module_source_path = os.path.normpath(directory + '/' + mod)
                target_path = os.path.normpath(directory)

                if module_source_path not in graph.nodes:
                    graph.add_node(module_source_path)

                if target_path not in graph.nodes:
                    graph.add_node(target_path)

                graph.add_edge(module_source_path, target_path)
    return graph


def _get_modules_for_file(directory: str, file_name: str) -> Tuple[str, Set[str]]:
    """
    Get the modules used in a terraform file
    :param directory: Directory where the file is in
    :param file_name: Name of file
    :return:
    """
    modules = set()
    with open(directory + '/' + file_name, 'r') as file:
        try:
            tf_info = hcl2.load(file)
            for module in tf_info.get('module', []):
                for module_config in module.values():
                    modules.add(os.path.normpath(module_config['source']))
        except Exception:
            print('Error while parsing file %s' % file.name)
            raise

    return directory, modules



"""Data classes to represent the wrapper config file"""
# TODO: convert these classes to dataclasses once we drop support for Python3.6
# pylint: disable=missing-docstring

from enum import Enum
from typing import Dict, Optional, List

import jsons


class EnvVarSource(Enum):
    SSM = 'ssm'
    TEXT = 'text'


class AbstractEnvVarConfig:
    def __init__(self, source: EnvVarSource):
        self.source = source


class SSMEnvVarConfig(AbstractEnvVarConfig):
    def __init__(self, path: str):
        super().__init__(EnvVarSource.SSM)
        self.path = path


class TextEnvVarConfig(AbstractEnvVarConfig):
    def __init__(self, value: str):
        super().__init__(EnvVarSource.TEXT)
        self.value = value


class S3BackendConfig:
    def __init__(
            self,
            bucket: str = None,
            region: str = None,
            dynamodb_table: str = None,
            role_arn: str = None,
    ):
        self.region = region
        self.bucket = bucket
        self.dynamodb_table = dynamodb_table
        self.role_arn = role_arn


class GCSBackendConfig:
    def __init__(self, bucket: str):
        self.bucket = bucket


class BackendsConfig:
    # pylint: disable=invalid-name
    def __init__(self, s3: Optional[S3BackendConfig] = None, gcs: Optional[GCSBackendConfig] = None):
        self.s3 = s3
        self.gcs = gcs


# pylint: disable=unused-argument
def env_var_deserializer(obj_dict, cls, **kwargs):
    """convert a dict to a subclass of AbstractEnvVarConfig"""
    if obj_dict['source'] == EnvVarSource.SSM.value:
        return SSMEnvVarConfig(obj_dict['path'])
    if obj_dict['source'] == EnvVarSource.TEXT.value:
        return TextEnvVarConfig(obj_dict['value'])

    raise RuntimeError('Invalid Source')


jsons.set_deserializer(env_var_deserializer, AbstractEnvVarConfig)


# pylint: disable=too-many-arguments
class WrapperConfig:
    def __init__(
            self,
            configure_backend: bool = True,
            pipeline_check: bool = True,
            backend_check: bool = True,
            plan_check: bool = True,
            envvars: Dict[str, AbstractEnvVarConfig] = None,
            backends: BackendsConfig = None,
            depends_on: List[str] = None,
            config: bool = True,
            apply_automatically: bool = True,
            plugins: Dict[str, str] = None
    ):
        self.configure_backend = configure_backend
        self.pipeline_check = pipeline_check
        self.backend_check = backend_check
        self.plan_check = plan_check
        self.envvars = envvars or {}
        self.backends = backends
        self.depends_on = depends_on
        self.config = config
        self.apply_automatically = apply_automatically
        self.plugins = plugins or {}



"""Module containing the Pipeline class"""
import concurrent.futures
import csv
from collections import defaultdict
from pathlib import Path
from typing import Iterable, Optional, List, DefaultDict

from terrawrap.models.pipeline_entry import PipelineEntry


class Pipeline:
    """Class for representing a pipeline."""

    def __init__(self, command: str, pipeline_path: str):
        """
        :param command: The Terraform command that this pipeline should execute.
        :param pipeline_path: Path to the Terrawrap pipeline file. Can be absolute or relative from where
        the script is executed.
        """
        self.command = command

        self.reverse_pipeline = command == 'destroy'

        if not pipeline_path.endswith(".csv"):
            raise RuntimeError("Config file '%s' doesn't appear to be a CSV file: Should end in .csv")

        with open(pipeline_path) as pipeline_file:
            reader = csv.DictReader(pipeline_file)
            # Lambda function is needed here because the argument to defaultdict needs to be a function that
            # returns an object.
            entries: DefaultDict[int, DefaultDict[str, List[PipelineEntry]]] = \
                defaultdict(lambda: defaultdict(list))

            for row in reader:
                entry = PipelineEntry(
                    path=row['directory'],
                    variables=row['variables'].split(' ') if row['variables'] else []
                )
                seq = int(row['seq'])
                path = Path(row['directory'])
                if not path.is_symlink():
                    entries[seq]['parallel'].append(entry)
                else:
                    entries[seq]['sequential'].append(entry)

        self.entries = entries

    def execute(self, num_parallel: int = 4, debug: bool = False, print_only_changes: bool = False):
        """
        Function for executing the pipeline. Will execute each sequence separately, with the entries inside
        each sequence being executed in parallel, up to the limit given in num_parallel.
        :param num_parallel: The number of pipeline entries to run in parallel.
        :param debug: True if Terraform debugging should be turned on.
        :param print_only_changes: True if only directories which contained changes should be printed.
        """
        for sequence in sorted(self.entries.keys(), reverse=self.reverse_pipeline):
            print("Executing sequence %s" % sequence)
            with concurrent.futures.ThreadPoolExecutor(max_workers=num_parallel) as executor:
                self._execute_entries(
                    command=self.command,
                    entries=self.entries[sequence]['parallel'],
                    debug=debug,
                    executor=executor,
                    print_only_changes=print_only_changes,
                )

            with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
                for entry in self.entries[sequence]['sequential']:
                    # It's very important that these sequential entries run init and then plan, and not all
                    # the inits and then all the plans, because the symlink directories might share the same
                    # real directories, and then running init multiple times in that directory will break
                    # future applies. So run init and then the command immediately.
                    self._execute_entries(
                        command=self.command,
                        entries=[entry],
                        debug=debug,
                        executor=executor,
                        print_only_changes=print_only_changes,
                    )

        print("Pipeline executed successfully.")

    def _execute_entries(
            self,
            entries: Iterable[PipelineEntry],
            executor: concurrent.futures.Executor,
            command: Optional[str] = None,
            debug: bool = False,
            print_only_changes: bool = False,
    ):
        """
        Convenience function for executing the given entries with the given command.
        :param entries: An iterable of Pipeline Entries to execute.
        :param executor: The Executor to use. See concurrent.futures.Executor.
        :param command: The Terraform command to execute. If not provided, defaults to the command for the
        pipeline.
        :param debug: True if Terraform debugging should be printed.
        :param print_only_changes: True if only directories which contained changes should be printed.
        """
        command = command or self.command
        futures_to_paths = {}
        failures = []

        for entry in entries:
            print("Executing %s %s ..." % (entry.path, command))
            future = executor.submit(entry.execute, command, debug=debug)
            futures_to_paths[future] = entry.path

        for future in concurrent.futures.as_completed(futures_to_paths):
            exit_code, stdout, changes_detected = future.result()
            path = futures_to_paths[future]

            if print_only_changes and not changes_detected:
                stdout = ["No changes detected.\n"]

            print("\nFinished executing %s %s ..." % (path, command))
            print("Output:\n\n%s\n" % "".join(stdout).strip())

            if exit_code != 0:
                failures.append(path)

        if failures:
            raise RuntimeError(
                "The follow pipeline entries failed with command '%s':\n%s" % (command, "\n".join(failures))
            )



"""Module for containing Pipeline Entries"""
import logging
import os
import tempfile
from typing import List, Tuple

from terrawrap.utils.cli import execute_command
from terrawrap.utils.config import find_wrapper_config_files, parse_wrapper_configs, resolve_envvars
from terrawrap.utils.path import get_absolute_path

logger = logging.getLogger(__name__)


class PipelineEntry:
    """Class representing a Pipeline Entry"""

    def __init__(self, path: str, variables: List[str]):
        """
        :param path: The path to the Terraform configuration files to execute.
        :param variables: Any additional variables to set alongside the Terraform command.
        """
        self.path = get_absolute_path(path=path)
        wrapper_config_files = find_wrapper_config_files(self.path)
        wrapper_config = parse_wrapper_configs(wrapper_config_files)
        self.envvars = resolve_envvars(wrapper_config.envvars)
        self.variables = variables

    # pylint: disable=too-many-locals
    def execute(self, operation: str, debug: bool = False) -> Tuple[int, List[str], bool]:
        """
        Function for executing this pipeline entry.
        :param operation: The Terraform operation to execute. IE: apply, plan
        :param debug: True if Terraform debug info should be printed.
        :return: A tuple of the exit code, output of the command, and whether changes were detected.
        """
        command_env = os.environ.copy()
        command_env.update(self.envvars)

        if debug:
            command_env["TF_LOG"] = "DEBUG"

        # pylint: disable=unused-variable
        plan_file, plan_file_name = tempfile.mkstemp(
            suffix=".tfplan"
        )

        # We're using --no-resolve-envvars here because we've already resolved the environment variables in
        # the constructor. We are then passing in those environment variables explicitly in the
        # execute_command call below.
        base_args = [
            "tf",
            "--no-resolve-envvars",
            self.path
        ]
        init_args = base_args + ["init"] + self.variables
        plan_args = base_args + [
            "plan",
            "-detailed-exitcode",
            "-out=%s" % plan_file_name
        ] + self.variables
        operation_args = base_args + [operation] + self.variables

        if operation in ["apply", "destroy"]:
            operation_args += ["-auto-approve"]

        init_exit_code, output = execute_command(
            init_args,
            print_output=False,
            capture_stderr=True,
            env=command_env,
        )

        if init_exit_code != 0:
            return init_exit_code, output, True

        # We need to plan first before we apply so we can actually see a plan... Thanks Hashicorp
        if operation in ["apply"]:
            plan_exit_code, plan_stdout = execute_command(
                plan_args,
                print_output=False,
                capture_stderr=True,
                env=command_env
            )

            output += ["\n"] + plan_stdout

            if plan_exit_code != 2:
                return (
                    plan_exit_code,
                    output,
                    False,
                )

            operation_args += [plan_file_name]

        if operation in ["plan"]:
            operation_args += ["--detailed-exitcode"]

        operation_exit_code, operation_stdout = execute_command(
            operation_args,
            print_output=False,
            capture_stderr=True,
            env=command_env
        )

        output += ["\n"] + operation_stdout

        changes_detected = True
        if operation in ["plan"]:
            if operation_exit_code == 2:
                # Set exit code to 0 so the command doesn't fail out
                operation_exit_code = 0
            elif operation_exit_code != 2:
                changes_detected = False

        return (
            operation_exit_code,
            output,
            changes_detected,
        )



"""Utility functions for working with Git"""
from typing import Set

import os
from git import Repo


def get_git_changed_files(path) -> Set[str]:
    """
    Compare HEAD of the current branch with master and return list of paths that changed
    Assumes this script is running from a git directory
    :return: List of paths that changed
    """
    repo = Repo(path, search_parent_directories=True)
    changed_files = set()
    for change in repo.commit('origin/master').diff(None):
        if not change.new_file:
            changed_files.add(os.path.abspath(change.a_path))
        if not change.deleted_file:
            changed_files.add(os.path.abspath(change.b_path))

    return changed_files


def get_git_root(path):
    """Get the git root directory for a given path"""
    git_repo = Repo(path, search_parent_directories=True)
    git_root = git_repo.git.rev_parse("--show-toplevel")
    return git_root



"""Utilities for working with collections"""
from typing import TypeVar, Dict

Type = TypeVar('Type')


def update(dict1: Dict, dict2: Dict) -> Dict:
    """
    Recursively updates the first provided dictionary with the keys and values from the second dictionary.
    Child dictionary and lists are merged, not replaced.
    :param dict1: The dictionary to merge into.
    :param dict2: The dictionary to merge.
    :return: A merged dictionary.
    """
    for key, value in dict2.items():
        if isinstance(value, dict):
            dict1[key] = update(dict1.get(key, {}), value)
        elif isinstance(value, list):
            original_value = dict1.get(key, [])
            original_value.extend(value)
            dict1[key] = original_value
        else:
            dict1[key] = value
    return dict1



"""setup.py controls the build, testing, and distribution of the egg"""
from __future__ import print_function

import re
import os.path

from setuptools import setup, find_packages


VERSION_REGEX = re.compile(r"""
    ^__version__\s=\s
    ['"](?P<version>.*?)['"]
""", re.MULTILINE | re.VERBOSE)

VERSION_FILE = os.path.join("terrawrap", "version.py")


def get_long_description():
    """Reads the long description from the README"""
    this_directory = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as file:
        return file.read()


def get_version():
    """Reads the version from the package"""
    with open(VERSION_FILE) as handle:
        lines = handle.read()
        result = VERSION_REGEX.search(lines)
        if result:
            return result.groupdict()["version"]
        raise ValueError("Unable to determine __version__")


def get_requirements():
    """Reads the installation requirements from requirements.pip"""
    with open("requirements.pip") as reqfile:
        return [line for line in reqfile.read().split("\n") if not line.startswith(('#', '-'))]


setup(
    name='terrawrap',
    python_requires='>=3.6.0',
    version=get_version(),
    description="Set of Python-based CLI tools for working with Terraform configurations",
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    # Get strings from http://www.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 4 - Beta',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    keywords='',
    author='Amplify Education',
    author_email='github@amplify.com',
    url='https://github.com/amplify-education/terrawrap',
    license='MIT',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=get_requirements(),
    test_suite='nose.collector',
    scripts=[
        'bin/tf',
        'bin/tf_apply',
        'bin/backend_check',
        'bin/pipeline_check',
        'bin/plan_check',
        'bin/visualize',
        'bin/graph_apply'
    ],
)
